package tree;

/**
 * Linked implementation of a binary search tree. The binary search tree
 * inherits the methods from the binary tree. The add and remove methods must
 * must be overridden so that they maintain the BST Property. The contains, get
 * and set methods are overridden to improve their performance by taking
 * advantage of the BST property. The inherited size and traversal methods work
 * well as they are.
 * 
 * @author Grant Braught
 * @author Dickinson College
 * @version March 10, 2016
 */
public class CS232LinkedBinarySearchTree<K extends Comparable<K>, V> extends
		CS232LinkedBinaryTree<K, V> {

	/*
	 * NOTE: We inherit the size and root fields, and the BTNode class from the
	 * LinkedBinaryTree class because they were declared as protected in that
	 * class.
	 */

	/**
	 * Construct an empty binary search tree.
	 */
	public CS232LinkedBinarySearchTree() {
		super();
	}

	/**
	 * Construct a binary search tree with a single node at the root.
	 * 
	 * @param key
	 *            the key for the root.
	 * @param value
	 *            the value for the root.
	 */
	public CS232LinkedBinarySearchTree(K key, V value) {
		super(key, value);
	}

	/**
	 * Construct a binary search tree using the provided keys and values. The
	 * key,value pairs are inserted into the tree in level order. If the
	 * resulting tree does not satisfy the BST Property, then an
	 * IllegalArgumentException is thrown.
	 * 
	 * @param keys
	 *            the keys.
	 * @param values
	 *            the values.
	 * @throws IllegalArgumentException
	 *             if the keys and values do not have the same length, the keys
	 *             or the values are empty, or the keys are not specified in an
	 *             order that results in a valid binary search tree.
	 */
	public CS232LinkedBinarySearchTree(K[] keys, V[] values) {
		super(keys, values);

		if (!checkBSTProperty()) {
			throw new IllegalArgumentException(
					"Key, Value pairs did not satisfy BST property.");
		}
	}

	/**
	 * {@inheritDoc}
	 */
	public boolean contains(K key) {
		return containsHelper(root, key);
	}

	/*
	 * Recursive helper method that checks if the key can be found in the
	 * subtree rooted at subTreeRoot.
	 */
	private boolean containsHelper(BTNode<K, V> subTreeRoot, K key) {
		if (subTreeRoot == null) {
			return false; // off the tree.
		} else if (subTreeRoot.key.equals(key)) {
			return true; // found it.
		} else if (key.compareTo(subTreeRoot.key) < 0) {
			/*
			 * The key we are looking for is less than the key at the
			 * subTreeRoot so if it is in the tree it will be in the left
			 * subtree.
			 */
			return containsHelper(subTreeRoot.left, key);
		} else {
			/*
			 * The key we are looking for is greater than or equal to the key at
			 * the subTreeRoot so if it is in the tree it will be in the right
			 * subtree.
			 */
			return containsHelper(subTreeRoot.right, key);
		}
	}

	/**
	 * {@inheritDoc}
	 */
	public V get(K key) {
		// Intentionally not implemented - see homework assignment.
		throw new UnsupportedOperationException("Not yet implemented");
	}

	/**
	 * {@inheritDoc}
	 */
	public void set(K key, V value) {
		// Intentionally not implemented - see homework assignment.
		throw new UnsupportedOperationException("Not yet implemented");
	}

	/**
	 * {@inheritDoc}
	 */
	public void add(K key, V value) {
		BTNode<K, V> nodeToAdd = getNewNode(key, value);
		addNodeToSubTree(root, nodeToAdd);
		size++;
	}

	/*
	 * Add the nodeToAdd to the subtree rooted at subTreeRoot.
	 */
	protected void addNodeToSubTree(BTNode<K, V> subTreeRoot,
			BTNode<K, V> nodeToAdd) {
		if (subTreeRoot == null) {
			root = nodeToAdd;
		}
		else if (nodeToAdd.key.compareTo(subTreeRoot.key) < 0) {
			/*
			 * Key of the new node is less than the key at subTreeRoot so we are
			 * going to add the new node into the left sub tree.
			 */
			if (subTreeRoot.left == null) {
				/*
				 * The left subtree is empty, so make the new node the left
				 * child of the subtree root.
				 */
				subTreeRoot.left = nodeToAdd;
				nodeToAdd.parent = subTreeRoot;
			} else {
				/*
				 * The left subtree is not empty, so add the new node to that
				 * sub tree.
				 */
				addNodeToSubTree(subTreeRoot.left, nodeToAdd);
			}
		} else {
			/*
			 * The key of the new node is greater than or equal to the key at
			 * the subTreeRoot so we are going to add the new node to the right
			 * sub tree. This is exactly the complement of the approach used
			 * above.
			 */
			if (subTreeRoot.right == null) {
				subTreeRoot.right = nodeToAdd;
				nodeToAdd.parent = subTreeRoot;
			} else {
				addNodeToSubTree(subTreeRoot.right, nodeToAdd);
			}
		}
	}

	/**
	 * {@inheritDoc}
	 */
	public V remove(K key) {
		// Intentionally not implemented - see homework assignment.
		throw new UnsupportedOperationException("Not yet implemented");
	}

	/**
	 * Determine whether the binary tree satisfies the binary search tree
	 * property (BSTP). That is, all values in the left subtree are strictly less
	 * than the root value. And all values in the right subtree are greater
	 * than or equal to the root value. And this property holds recursively
	 * for every other node in the tree.
	 * 
	 * @return true if the tree satisfies the binary search tree property,
	 *         and false otherwise.
	 */
	public boolean checkBSTProperty() {
		return checkBSTPropertyHelp(root, null, null);
	}

	/**
	 * Determine whether the binary tree with the given root node satisfies
	 * the binary search tree property (BSTP) AND all keys in the tree are
	 * strictly less than the 'high' parameter AND all keys in the tree are
	 * greater than or equal to the 'low' parameter. If 'high' is null,
	 * there is no maximum on the keys; if 'low' is null, there is no
	 * minimum on the keys. Note that our definition of the binary search
	 * tree property insists that left children are strictly smaller, while
	 * right children can be equal or strictly greater.
	 * 
	 * This implementation is adapted from OpenDSA textbook materials
	 * (https://opendsa-server.cs.vt.edu/). Specifically, this is from
	 * section 15 of the Binary tree chapter, in e.g.
	 * https://opendsa-server.cs.vt.edu/OpenDSA/Books/CS2/html/
	 * 
	 * Note that the OpenDSA implementation assumes that any equal keys are
	 * in the left subtree, whereas our implementation assumes they are in
	 * the right subtree. This leads to certain differences in the
	 * algorithm, but the principles are the same.
	 * 
	 * @param subTreeRoot The root of the subtree to be checked
	 * @param low         The inclusive minimum for all keys in the tree
	 * @param high        The exclusive maximum for all keys in the tree
	 * @return true if BSTP holds and all keys are at least 'low' and all
	 *         keys are less than 'high'
	 * 
	 */
	public boolean checkBSTPropertyHelp(BTNode<K, V> subTreeRoot, K low, K high) {
		if (subTreeRoot == null) {
			// Empty subtree
			return true;
		}
		K k = subTreeRoot.key;
		if (high != null && k.compareTo(high) >= 0) {
			return false;
		}
		if (low != null && k.compareTo(low) < 0) {
			return false;
		}
		return checkBSTPropertyHelp(subTreeRoot.left, low, k)
				&& checkBSTPropertyHelp(subTreeRoot.right, k, high);
	}
}
